// Contains a random integer ID used for HTTP subscriptions
var backendChannelSubscriptionID = null;

// Contains an integer timestamp that will be the base value for
// all the other timestamps which will be used for message IDs
var backendChannelBaseTimestamp = null;

// Contains response functions for associated message IDs
var backendChannelResponseFunctions = {};


/* Get a property string value from property tree.
 * @param propPath property path
 * @return property value, or null if not found */
function getPropertyString(propPath)
{
	var url ='/json/property/getString';
	var path = escape(propPath);
	var retval = dss.ajaxSyncRequest(url, { path: path } );
	var obj = eval('(' + retval + ')');
	if (obj.ok)
		return obj.result.value;
	else
		return null;
};


/* Converts a string containing an HTML hexadecimal color value to an array of three R/G/B values.
 *
 * @param hex HTML color value (format: #RRGGBB)
 * @return array of the three R/G/B values, in the 0..255 range */
function hex2rgb(hexColor)
{
	// slightly modified version of http://stackoverflow.com/a/11508164
	var h = hexColor.replace('#', ''); // strip the hash at the beginning
	var bigint = null;
	return [((bigint = parseInt(h, 16)) >> 16) & 255, (bigint >> 8 & 255), (bigint & 255)];
}


/* Converts an array of three R/G/B values to a string containing an HTML hexadecimal color value.
 *
 * @param rgb array of the three R/G/B values, in the 0..255 range
 * @return HTML color value (format: #RRGGBB) */
function rgb2hex(rgb)
{
	// slightly modified version of http://stackoverflow.com/a/19765382
	var rgb = rgb[2] | (rgb[1] << 8) | (rgb[0] << 16);
	return '#' + (0x1000000 + rgb).toString(16).slice(1);
}


function percent2Brightness(percent)
{
	return Math.round(percent * 255 / 100);
}

function brightness2Percent(percent)
{
	return Math.round(percent * 100 / 255);
}


/* Converts a mired temperature value to the Kelvin range
 * @param mired Mired temperature value
 * @return Color temperature in Kelvin */
function mired2Kelvin(mired)
{
	var kelvin = Math.floor(1000000 / mired);
	return Math.min(6500, Math.max(2000, kelvin));
}


/* Converts a Kelvin color temperature value to Mired units
 * @param kelvin Kelvin color temperature
 * @return Color temperature in Mired units */
function kelvin2Mired(kelvin)
{
	var mired = Math.floor(1000000 / kelvin);
	return Math.min(500, Math.max(153, mired));
}


/* RGB -> HSV conversion function.
 *
 * @param rgb Array containing the 3 integer values: red, green, blue, in the [0..255] range
 * @return hsv Array containing the 3 decimal values: hue, saturation, value, in the [0..1] range */
function rgb2hsv(rgb)
{
	var r = rgb[0] / 255.0;
	var g = rgb[1] / 255.0;
	var b = rgb[2] / 255.0;

	var minValue = Math.min(r, g, b);
	var maxValue = Math.max(r, g, b);
	var delta = maxValue - minValue;

	var hue = 0;
	var saturation = 0;
	var value = maxValue;

	if (maxValue != 0)
		saturation = delta / maxValue;

	if (delta != 0)
	{

		if (r === maxValue)
			hue = 0 + (g - b) / delta;
		else if (g === maxValue)
			hue = 2 + (b - r) / delta;
		else
			hue = 4 + (r - g) / delta;

		if (hue < 0)
			hue += 6;

		hue = hue / 6.0;
	}

	return [hue, saturation, value];
}


/* HSV -> RGB conversion function.
 *
 * @param hsv Array containing the 3 decimal values: hue, saturation, value, in the [0..1] range
 * @return rgb Array containing the 3 integer values: red, green, blue, in the [0..255] range */
function hsv2rgb(hsv)
{
	var H = hsv[0] * 6.0;
	var S = hsv[1];
	var V = hsv[2];

	var h = Math.floor(H);
	var f = H - h;
	var p = V * (1.0 - S);
	var q = V * (1.0 - S * f);
	var t = V * (1.0 - S * (1.0 - f));

	var rgb = [];

	switch (h)
	{
		case 0:
		case 6: rgb = [ V, t, p ]; break;
		case 1: rgb = [ q, V, p ]; break;
		case 2: rgb = [ p, V, t ]; break;
		case 3: rgb = [ p, q, V ]; break;
		case 4: rgb = [ t, p, V ]; break;
		case 5: rgb = [ V, p, q ]; break;
		default: rgb = [ 0, 0, 0 ];
	}

	rgb[0] = Math.floor(rgb[0] * 255.0);
	rgb[1] = Math.floor(rgb[1] * 255.0);
	rgb[2] = Math.floor(rgb[2] * 255.0);

	return rgb;
}


/* Return a name for the given scene including any custom name.
 *
 * Scenes can have custom names. This function returns a name
 * that is either the regular scene name or the regular name + the
 * custom one.
 *
 * @param sceneID ID of the scene to get the name for
 * @param groupID Group ID for the scene
 * @param zoneID Zone ID for the scene
 * @return The full name including the regular and the custom one */
function getCustomNameForScene(sceneID, groupID, zoneID)
{
	var scenes = dss.staticDataModel.getScenes(zoneID, groupID);

	for (var i = 0; i < scenes.length; ++i)
	{
		var scene = scenes[i];
		if (scene.id === sceneID)
		{
			if (scene.name === scene.defaultName)
				return scene.defaultName;
			else
				return scene.defaultName + ' - ' + scene.name;
		}
	}

	return dss.staticDataModel.getNameForScene(zoneID, groupID, sceneID);
}


/* Issues an event subscription if not already done before. */
function resubscribeBackendChannel()
{
	if (backendChannelSubscriptionID !== null)
		return;

	try
	{
		/* Get a random ID */
		backendChannelSubscriptionID = Math.floor(Math.random() * 1000000);
		console.log("Initializing backend channel HTTP subscription with ID " + String(backendChannelSubscriptionID));

		/* Do the actual subscription */
		dss.ajaxSyncRequest(
			'/json/event/subscribe',
			{
				subscriptionID : backendChannelSubscriptionID,
				name : 'huecontrollerbackendChannel'
			}
		);
	}
	catch (e)
	{
		console.error("could not initialize HTTP backend communication channel");
	}
}


/* Returns a timestamp for the current time.
 * @return Integer containing the current time, in microseconds */
function getCurrentTimestamp()
{
	var p = window.performance || {};

	if (p.now)
		return Math.floor(p.now() * 1e11);
	else if (p.mozNow)
		return Math.floor(p.mozNow() * 1e11);
	else if (p.msNow)
		return Math.floor(p.msNow() * 1e11);
	else if (p.oNow)
		return Math.floor(p.oNow() * 1e11);
	else if (p.webkitNow)
		return Math.floor(p.webkitNow() * 1e11);
	else
		return Date.now();
}


/* Generates a unique message ID with the given name.
 * The message ID is a combination of the name, the current time,
 * and a random integer number.
 * The random number is added to handle the rare case where a
 * second ID is requested almost immediately after a first.
 * Since the timestamp granularity might be worse than 1 microsecond,
 * the random number helps to maintain uniqueness.
 *
 * @return The unique ID string */
function generateMsgId(name)
{
	var timestamp = Date.now(); // in milliseconds;
	var rndnum = Math.floor(Math.random() * 1000);

	/* To make debugging a bit easier, make sure
	 * the timestamp of the very first message ID is 0 */
	if (backendChannelBaseTimestamp === null)
		backendChannelBaseTimestamp = timestamp;
	timestamp -= backendChannelBaseTimestamp;

	return String(timestamp) + '-' + String(rndnum) + '-' + name;
}


/* Performs an asynchronous backend request.
 *
 * If no response shall be handled, set responseFunction to null.
 *
 * @param messageName name of the message to send to the backend.
 * @param payload object containing the payload to send.
 * @param responseFunction function to be called when a response is received */
function sendToBackend(messageName, messagePayload, responseFunction)
{
	/* if the subscription ID is null at this point, then the backend
	 * channel loop has not been started yet -> start it */
	if (backendChannelSubscriptionID === null)
	{
		console.log("sendToBackend: starting backend channel loop");
		backendChannelLoop();
	}

	var encodedPayload;
	try
	{
		encodedPayload = Ext.JSON.encode(messagePayload);
	}
	catch (err)
	{
		console.error("Error while encoding message payload as JSON: " + String(err));
		return;
	}

	var msgid = generateMsgId(messageName);

	/* Store the response function in a table */
	if (responseFunction)
		backendChannelResponseFunctions[msgid] = responseFunction;
	console.log('sendToBackend:   message id: "' + msgid + '"  name: "' + messageName + '"  JSON-encoded payload: ' + encodedPayload);

	dss.raiseEvent('huecontroller', { action: messageName, msgid: msgid, payload: encodedPayload });
}


/* Backend channel loop for long-polling based communication.
 *
 * Communication with the backend works asynchronously. Messages to
 * the backend are sent in sendToBackend() using raiseEvent(). Messages
 * from the backend arrive through this backend channel. It uses the
 * long-polling technique: send a GET request, the backend can then put
 * messages inside the GET response. If the backend does not need to
 * send messages for a while, the GET request simply timeouts, and the
 * frontend issues a new one. */
function backendChannelLoop()
{
	/* (Re)subscribe if necessary */
	resubscribeBackendChannel();

	dss.ajaxAssyncRequestWithoutAnswer(
		'/json/event/get',
		{
			subscriptionID : backendChannelSubscriptionID,
			timeout : 5 * 1000
		},
		function(response)
		{
			console.info("backendChannelLoop: received response: " + String(response));

			/* Parse response */
			var decResponse;
			try
			{
				decResponse = Ext.JSON.decode(response);
			}
			catch (err)
			{
				/* Invalid JSON? Report error, and continue the loop */

				console.error("Error while parsing message payload: " + String(err));

				backendChannelSubscriptionID = null;
				resubscribeBackendChannel();
				/* continue the async loop */
				backendChannelLoop();

				return;
			}

			if (decResponse.ok === false)
			{
				/* Print error and resubscribe to maintain this channel */

				if (decResponse.message === "Subscription id not found!")
					console.log("Subscription ID lost - resubscribing new ID and retransmitting request");
				else if (decResponse.message === "Invalid session!")
					console.log("Session invalid - resubscribing new ID and retransmitting request");
				else
					console.error(String(decResponse.message));

				backendChannelSubscriptionID = null;
				resubscribeBackendChannel();
				/* continue the async loop */
				backendChannelLoop();

				return;
			}

			if (decResponse.result)
			{
				/* There is actual data to process */

				for (var i = 0; i < decResponse.result.events.length; ++i)
				{
					var props = decResponse.result.events[i].properties;
					var msgid = String(props.msgid);

					/* If there is a response function that corresponds
					 * to the present message ID, fetch and call it */
					var responseFunction = backendChannelResponseFunctions[msgid];
					if (responseFunction)
					{
						console.info('backendChannelLoop: response function found for message with ID "' + msgid + '"');
						/* clear the function from the table to catch error
						 * situations where the same msgid is used twice */
						delete backendChannelResponseFunctions[msgid];
						responseFunction(props);
					}
					else
					{
						console.info('backendChannelLoop: no response function found for message with ID "' + msgid + '"');
					}
				}
			}
			else
				console.warn('backendChannelLoop: response result field is not present');

			/* continue the async loop */
			backendChannelLoop();
		}
	);
}

